use crate::context::{AnalyzerContext, CallType, FunctionBody};
use crate::db::{Analysis, AnalyzerDb};
use crate::errors::TypeError;
use crate::namespace::items::{DepGraph, DepGraphWrapper, DepLocality, FunctionId, Item, TypeDef};
use crate::namespace::scopes::{BlockScope, BlockScopeType, FunctionScope, ItemScope};
use crate::namespace::types::{self, Contract, CtxDecl, SelfDecl, Struct, Type};
use crate::traversal::functions::traverse_statements;
use crate::traversal::types::type_desc;
use fe_common::diagnostics::Label;
use fe_parser::ast;
use fe_parser::node::Node;
use if_chain::if_chain;
use std::collections::HashMap;
use std::rc::Rc;

/// Gather context information for a function definition and check for type
/// errors. Does not inspect the function body.
pub fn function_signature(
    db: &dyn AnalyzerDb,
    function: FunctionId,
) -> Analysis<Rc<types::FunctionSignature>> {
    let node = &function.data(db).ast;
    let def = &node.kind;

    let mut scope = ItemScope::new(db, function.module(db));
    let fn_parent = function.class(db);

    let mut self_decl = None;
    let mut ctx_decl = None;
    let mut names = HashMap::new();
    let mut labels = HashMap::new();

    let params = def
        .args
        .iter()
        .enumerate()
        .filter_map(|(index, arg)| match &arg.kind {
            ast::FunctionArg::Self_ => {
                if fn_parent.is_none() {
                    scope.error(
                        "`self` can only be used in contract or struct functions",
                        arg.span,
                        "not allowed in functions defined outside of a contract or struct",
                    );
                } else {
                    self_decl = Some(SelfDecl::Mutable);
                    if index != 0 {
                        scope.error(
                            "`self` is not the first parameter",
                            arg.span,
                            "`self` may only be used as the first parameter",
                        );
                    }
                }
                None
            }
            ast::FunctionArg::Regular(reg) => {
                let typ = type_desc(&mut scope, &reg.typ).and_then(|typ| match typ {
                    typ if typ.has_fixed_size() => Ok(typ),
                    _ => Err(TypeError::new(scope.error(
                        "function parameter types must have fixed size",
                        reg.typ.span,
                        "`Map` type can't be used as a function parameter",
                    ))),
                });

                if let Some(context_type) = scope.get_context_type() {
                    if typ == Ok(Type::Struct(context_type)) {
                        if arg.name() != "ctx" {
                            scope.error(
                                "invalid `Context` instance name",
                                arg.span,
                                "instances of `Context` must be named `ctx`",
                            );
                        } else if !function.parent(db).is_contract() {
                            scope.error(
                                "`ctx` cannot be passed into pure functions",
                                arg.span,
                                "`ctx` can only be passed into contract functions",
                            );
                        } else if self_decl.is_some() && index != 1 {
                            scope.error(
                                "invalid parameter order",
                                arg.span,
                                "`ctx: Context` must be placed after the `self` parameter",
                            );
                        } else if self_decl.is_none() && index != 0 {
                            scope.error(
                                "invalid parameter order",
                                arg.span,
                                "`ctx: Context` must be the first parameter",
                            );
                        } else {
                            ctx_decl = Some(CtxDecl::Mutable);
                        }
                    }
                }

                if let Some(label) = &reg.label {
                    if_chain! {
                        if label.kind != "_";
                        if let Some(dup_idx) = labels.get(&label.kind);
                        then {
                            let dup_arg: &Node<ast::FunctionArg> = &def.args[*dup_idx];
                            scope.fancy_error(
                                &format!("duplicate parameter labels in function `{}`", def.name.kind),
                                vec![
                                    Label::primary(dup_arg.span, "the label `{}` was first used here"),
                                    Label::primary(label.span, "label `{}` used again here"),
                                ], vec![]);
                            return None;
                        } else {
                            labels.insert(&label.kind, index);
                        }
                    }
                }

                if let Ok(Some(named_item)) = scope.resolve_name(&reg.name.kind, reg.name.span) {
                    scope.name_conflict_error(
                        "function parameter",
                        &reg.name.kind,
                        &named_item,
                        named_item.name_span(db),
                        reg.name.span,
                    );
                    None
                } else if let Some(dup_idx) = names.get(&reg.name.kind) {
                    let dup_arg: &Node<ast::FunctionArg> = &def.args[*dup_idx];
                    scope.duplicate_name_error(
                        &format!("duplicate parameter names in function `{}`", def.name.kind),
                        &reg.name.kind,
                        dup_arg.span,
                        arg.span,
                    );
                    None
                } else {
                    names.insert(&reg.name.kind, index);
                    Some(types::FunctionParam::new(
                        reg.label.as_ref().map(|s| s.kind.as_str()),
                        &reg.name.kind,
                        typ,
                    ))
                }
            }
        })
        .collect();

    let return_type = def
        .return_type
        .as_ref()
        .map(|type_node| {
            let fn_name = &def.name.kind;
            if fn_name == "__init__" || fn_name == "__call__" {
                // `__init__` and `__call__` must not return any type other than `()`.
                if type_node.kind != ast::TypeDesc::Unit {
                    scope.fancy_error(
                        &format!("`{}` function has incorrect return type", fn_name),
                        vec![Label::primary(type_node.span, "return type should be `()`")],
                        vec![
                            "Hint: Remove the return type specification.".to_string(),
                            format!("Example: `pub fn {}():`", fn_name),
                        ],
                    );
                }
                Ok(Type::unit())
            } else {

                match type_desc(&mut scope, type_node)? {
                    Type::Struct(val) if val.id.has_complex_fields(db) && function.is_public(db) => {
                        scope.not_yet_implemented("structs with complex fields can't be returned from public functions yet", type_node.span);
                        Ok(Type::Struct(val))
                    }
                    typ if typ.has_fixed_size() => Ok(typ),
                    _ => Err(TypeError::new(scope.error(
                        "function return type must have a fixed size",
                        type_node.span,
                        "this can't be returned from a function",
                    ))),
                }
            }
        })
        .unwrap_or_else(|| Ok(Type::unit()));

    Analysis {
        value: Rc::new(types::FunctionSignature {
            self_decl,
            ctx_decl,
            params,
            return_type,
        }),
        diagnostics: scope.diagnostics.take().into(),
    }
}

/// Gather context information for a function body and check for type errors.
pub fn function_body(db: &dyn AnalyzerDb, function: FunctionId) -> Analysis<Rc<FunctionBody>> {
    let def = &function.data(db).ast.kind;
    let scope = FunctionScope::new(db, function);

    // If the return type is unit, explicit return or no return (implicit) is valid,
    // so no scanning is necessary.
    // If the return type is anything else, we need to ensure that all code paths
    // return or revert.
    if let Ok(return_type) = &function.signature(db).return_type {
        if !return_type.is_unit() && !all_paths_return_or_revert(&def.body) {
            scope.fancy_error(
                "function body is missing a return or revert statement",
                vec![
                    Label::primary(
                        def.name.span,
                        "all paths of this function must `return` or `revert`",
                    ),
                    Label::secondary(
                        def.return_type.as_ref().unwrap().span,
                        format!("expected function to return `{}`", return_type),
                    ),
                ],
                vec![],
            );
        }
    }

    let mut block_scope = BlockScope::new(
        &scope,
        if function.is_unsafe(db) {
            BlockScopeType::Unsafe
        } else {
            BlockScopeType::Function
        },
    );

    // If `traverse_statements` fails, we can be confident that a diagnostic
    // has been emitted, either while analyzing this fn body or while analyzing
    // a type or fn used in this fn body, because of the `DiagnosticVoucher`
    // system. (See the definition of `FatalError`)
    let _ = traverse_statements(&mut block_scope, &def.body);
    Analysis {
        value: Rc::new(scope.body.into_inner()),
        diagnostics: scope.diagnostics.into_inner().into(),
    }
}

fn all_paths_return_or_revert(block: &[Node<ast::FuncStmt>]) -> bool {
    for statement in block.iter().rev() {
        match &statement.kind {
            ast::FuncStmt::Return { .. } | ast::FuncStmt::Revert { .. } => return true,
            ast::FuncStmt::If {
                test: _,
                body,
                or_else,
            } => {
                let body_returns = all_paths_return_or_revert(body);
                let or_else_returns = all_paths_return_or_revert(or_else);
                if body_returns && or_else_returns {
                    return true;
                }
            }
            ast::FuncStmt::Unsafe(body) => {
                if all_paths_return_or_revert(body) {
                    return true;
                }
            }
            _ => {}
        }
    }

    false
}

pub fn function_dependency_graph(db: &dyn AnalyzerDb, function: FunctionId) -> DepGraphWrapper {
    let root = Item::Function(function);

    // Edges to direct dependencies.
    let mut directs = vec![];

    let sig = function.signature(db);
    directs.extend(
        sig.return_type
            .clone()
            .into_iter()
            .chain(sig.params.iter().filter_map(|param| param.typ.clone().ok()))
            .filter_map(|typ| match typ {
                Type::Contract(Contract { id, .. }) => {
                    // Contract types that are taken as (non-self) args or returned are "external",
                    // meaning that they're addresses of other contracts, so we don't have direct
                    // access to their fields, etc.
                    Some((
                        root,
                        Item::Type(TypeDef::Contract(id)),
                        DepLocality::External,
                    ))
                }
                Type::Struct(Struct { id, .. }) => {
                    Some((root, Item::Type(TypeDef::Struct(id)), DepLocality::Local))
                }
                _ => None,
            }),
    );
    // A function that takes `self` depends on the type of `self`, so that any
    // relevant struct getters/setters are included when compiling.
    if let Some(class) = function.class(db) {
        directs.push((root, class.as_item(), DepLocality::Local));
    }

    let body = function.body(db);
    for calltype in body.calls.values() {
        match calltype {
            CallType::Pure(function) | CallType::AssociatedFunction { function, .. } => {
                directs.push((root, Item::Function(*function), DepLocality::Local));
            }
            CallType::ValueMethod { class, method, .. } => {
                // Including the "class" type here is probably redundant; the type will
                // also be part of the fn sig, or some type decl, or some create/create2 call,
                // or...
                directs.push((root, class.as_item(), DepLocality::Local));
                directs.push((root, Item::Function(*method), DepLocality::Local));
            }
            CallType::External { contract, function } => {
                directs.push((root, Item::Function(*function), DepLocality::External));
                // Probably redundant:
                directs.push((
                    root,
                    Item::Type(TypeDef::Contract(*contract)),
                    DepLocality::External,
                ));
            }
            CallType::TypeConstructor(Type::Struct(Struct { id, .. })) => {
                directs.push((root, Item::Type(TypeDef::Struct(*id)), DepLocality::Local));
            }
            CallType::TypeConstructor(Type::Contract(Contract { id, .. })) => {
                directs.push((
                    root,
                    Item::Type(TypeDef::Contract(*id)),
                    DepLocality::External,
                ));
            }
            CallType::TypeConstructor(_) => {}
            CallType::BuiltinAssociatedFunction { contract, .. } => {
                // create/create2 call. The contract type is "external" for dependency graph
                // purposes.
                directs.push((
                    root,
                    Item::Type(TypeDef::Contract(*contract)),
                    DepLocality::External,
                ));
            }
            // Builtin functions aren't part of the dependency graph yet.
            CallType::BuiltinFunction(_)
            | CallType::Intrinsic(_)
            | CallType::BuiltinValueMethod { .. } => {}
        }
    }

    directs.extend(
        body.emits
            .values()
            .map(|event| (root, Item::Event(*event), DepLocality::Local)),
    );
    directs.extend(body.var_decl_types.values().filter_map(|typ| match typ {
        Type::Contract(Contract { id, .. }) => Some((
            root,
            Item::Type(TypeDef::Contract(*id)),
            DepLocality::External,
        )),
        Type::Struct(Struct { id, .. }) => {
            Some((root, Item::Type(TypeDef::Struct(*id)), DepLocality::Local))
        }
        _ => None,
    }));

    let mut graph = DepGraph::from_edges(directs.iter());
    for (_, item, _) in directs {
        if let Some(subgraph) = item.dependency_graph(db) {
            graph.extend(subgraph.all_edges())
        }
    }
    DepGraphWrapper(Rc::new(graph))
}

pub fn function_dependency_graph_cycle(
    _db: &dyn AnalyzerDb,
    _cycle: &[String],
    _function: &FunctionId,
) -> DepGraphWrapper {
    DepGraphWrapper(Rc::new(DepGraph::new()))
}
